-module(lazy).
-compile(export_all).
-compile({no_auto_import,[tl/1,hd/1]}).
-author('hanak@emt.bme.hu, kapolnai@iit.bme.hu').
-vsn('$LastChangedDate: 2019-10-30 07:32:13 +0100 (sze, 30 okt 2019) $$').

-type lazy_list() :: nil | {any(), fun(() -> lazy_list())}.
% Ez a lista félig lusta: a fej mindig kiértékelődik, a farok lusta.
% A teljesen lusta lista kicsit komplikáltabb, mert nem tudni, hogy
% üres vagy sem.

-spec infseq(N::integer()) -> lazy_list().
infseq(N) ->
    {N, fun() -> infseq(N+1) end}.

-spec seq(M::integer(), N::integer()) -> lazy_list().
seq(M, N) when M =< N ->
    {M, fun() -> seq(M+1, N) end};
seq(_, _) ->
    nil.

-spec cons(list()) -> lazy_list().
cons([]) ->
    nil;
cons([H|T]) ->
    {H, fun() -> cons(T) end}.

-spec take(L::lazy_list(), N::integer()) -> list().
take(_,0) ->
    [];
take(nil,_) ->
    [];
take({H,_}, 1) ->  % optimalizáció: hogy ne kelljen a farkát kiértékelni T()
    [H];
take({H,T}, N) ->
    [H|take(T(), N-1)].

% Szószátyár take nyomkövetéshez.
take2(_L,0) ->
    io:format("hivas: take(~w, 0)~n", [_L]), % nyomkoveteshez
    [];
take2({H,_T},1) ->
    io:format("hivas: take(~w, ~w)~n", [[H|_T], 1]), % nyomkoveteshez
    [H];
take2(nil,_N) ->
    io:format("hivas: take([], ~w)~n", [_N]), % nyomkoveteshez
    [];
take2({H,T},N) ->
    io:format("hivas: take(~w, ~w)~n", [[H|T], N]), % nyomkoveteshez
    [H|take2(T(), N-1)].

-spec sum(lazy_list()) -> number().
sum(L) ->
    sum(L, 0).

-spec sum(lazy_list(), number()) -> number().
% Csak véges lusta listákra!
sum(nil, X) ->
    X;
sum({H,T}, X) ->
    sum(T(), H+X).

-spec map(fun((any()) -> any()), lazy_list()) -> lazy_list().
map(_, nil) ->
    nil;
map(F, {H,T}) ->
    {F(H), fun() -> map(F, T()) end}.

-spec filter(fun((any()) -> boolean()), lazy_list()) -> lazy_list().
% Kicsit mohó, az eredménylista fejéig kiértékeli a listát.
filter(_, nil) ->
    nil;
filter(P, {H,T}) ->
    case P(H) of
        true ->
            {H, fun() -> filter(P, T()) end};
        false ->
            filter(P, T()) % Megkeressük az eredmény fejét.
    end.

-spec append(lazy_list(), lazy_list()) -> lazy_list().
append(nil, L2) ->
    L2;
append({H,T}, L2) ->
    {H, fun() -> append(T(), L2) end}.

% Szószátyár append nyomkövetéshez.
append2(nil, L2) ->
    io:format("    Hivas: append([], ~w)~n", [(L2)]), % nyomkoveteshez
    L2;
append2({H,T}, L2) ->
    io:format("    Hivas: append(~w, ~w)~n", [([H|T]), (L2)]), % nyomkoveteshez
    {H, fun() -> append2(T(), L2) end}.

-spec fibs(integer(), integer()) -> lazy_list().
fibs(Cur, Next) ->
    {Cur, fun() -> fibs(Next, Cur+Next) end}.

-spec fib(integer()) -> integer().
fib(N) ->
    nth(fibs(0,1), N+1).

-spec sift(Prime::integer(), L::lazy_list()) -> L2::lazy_list().
% Az L lista azon elemeinek listája L2, melyek nem oszthatók Prime-mal.
sift(Prime, L) ->
    filter(fun(N) -> N rem Prime =/= 0 end, L).

-spec sieve(L1::lazy_list()) -> L2::lazy_list().
% Az L2 az L1 végtelen lista szitáltja (üres listára hibát ad).
sieve({H,T}) ->
    {H, fun() -> sieve(sift(H, T())) end}.

-spec qsort(lazy_list()) -> lazy_list().
% Az append alkalmazása: ha csak a lista elejére kíváncsi
% a felhasználó, csak a lista elejét rendezi.
qsort(nil) ->
    nil;
qsort({Pivot, Xs}) ->
    append(qsort(filter(fun(X) -> X < Pivot end, Xs())),
           {Pivot, fun() -> qsort(filter(fun(X) -> X >= Pivot end, Xs())) end}).

% Szószátyár qsort nyomkövetéshez.
qsort2(nil) ->
    nil;
qsort2({Pivot,Xs}) ->
    io:format("Hivas: qsort2(~w)~n", [take({Pivot,Xs}, 100)]), % nyomkoveteshez
    Low = fun(X) -> X < Pivot end,
    High = fun(X) -> X >= Pivot end,
    append2(qsort2(filter(Low, Xs())),
           {Pivot, fun() -> qsort2(filter(High, Xs())) end}).

-spec tl(lazy_list()) -> lazy_list().
tl({_,T}) ->
    T().

-spec hd(lazy_list()) -> any().
hd({H,_T}) ->
    H.

-spec nth(L::lazy_list(), N::integer()) -> X::any().
%% Az L lusta lista N-edik eleme X (számozás 1-től).
nth({H,_T}, 1) ->
    H;
nth({_H,T}, N) ->
    nth(T(), N-1).

-spec nth2(L::lazy_list(), N::integer()) -> {ok, X::any()} | error.
%% Az L lusta lista N-edik eleme X (számozás 1-től). A visszatérési
%% érték az 'error' atom, ha az L lista üres, vagy ha N < 1.
nth2({H,_T}, 1) ->
    {ok, H};
nth2({_H,T}, N) when N > 1 ->
    nth2(T(), N-1);
nth2(_, _) ->
    error.

test(1) ->
    LL = {1, fun() -> {4, fun() -> {9,fun() -> nil end} end} end},
    [
     sum(seq(1,100000)) =:= 5000050000,
     sum(cons(lists:seq(1,100000))) =:= 5000050000,
     take(LL, 2) =:= [1,4],
     take(LL, 3) =:= [1,4,9],
     take(cons([1,4,9]), 4) =:= [1,4,9],
     take(seq(1,3),10) =:= [1,2,3],
     take(infseq(1),10) =:= [1,2,3,4,5,6,7,8,9,10],
     hd(tl(tl(infseq(1)))) =:= 3,
     take(tl(tl(infseq(1))),1) =:= [3],
     take(fibs(0,1),10+1) =:= [0,1,1,2,3,5,8,13,21,34,55],
     nth(fibs(0,1), 100+1) =:= 354224848179261915075,
     fib(100) =:= 354224848179261915075,
     nth2(fibs(0,1), 100+1) =:= {ok,354224848179261915075},
     nth2(nil, 5) =:= error,
     nth2(fibs(0,1), 0) =:= error,
     take(map(fun(X) -> X rem 2 =:= 0 end, infseq(3)), 3) =:= [false,true,false],
     take(filter(fun(X) -> X rem 2 =:= 0 end, infseq(3)), 3) =:= [4,6,8],
     take(filter(fun(X) -> X rem 2 =:= 1 end, seq(3,6)), 3) =:= [3,5],
     [take(append(cons("ab"),cons("12")), I) || I <- lists:seq(0,4)] =:=
	 [[],"a","ab","ab1","ab12"],
     take(sieve(infseq(2)), 10) =:= [2,3,5,7,11,13,17,19,23,29]
    ];
test(2) ->
    [
     take(qsort(cons([5,3,6,8,1,7])), 2) =:= [1,3],
     take(qsort(cons([5,3,6,8,1,7])), 5) =:= [1,3,5,6,7],
     io:format("~nTeszthívás: ~s~n", ["take(qsort2(cons([5,3,6,8,1,7])), 2)"]),
     take(qsort2(cons([5,3,6,8,1,7])), 2) =:= [1,3],
     io:format("~nTeszthívás: ~s~n", ["take(qsort2(cons([5,3,6,8,1,7])), 5)"]),
     take(qsort2(cons([5,3,6,8,1,7])), 5) =:= [1,3,5,6,7]
    ].

-spec print_mem_usage(string()) -> {integer(), non_neg_integer()}.
% Durva becslés: kiírja a függvényhívás előtt és után használt memóriát.
print_mem_usage(Msg) ->
    {memory,M1} = erlang:process_info(self(), memory),
    {_,Time} = statistics(runtime),
    io:format("~16s: ~3w M  |  ~.3f s~n", [Msg, M = M1 div 1024 div 1024, Time / 1000]),
    {M,Time}.

memtest(1) ->
    erlang:garbage_collect(),timer:sleep(1000),statistics(runtime),
    sum(seq(1,10000000)),
    print_mem_usage("seq o sum");
memtest(2) ->
    erlang:garbage_collect(),timer:sleep(1000),statistics(runtime),
    L = lists:seq(1,10000000),
    print_mem_usage("lists:seq"),
    lists:sum(L),
    print_mem_usage("lists:sum").
